@@ -94,6 +94,21 @@ class TestCase(BaseTestCase):
 
         return (queue1_path, queue2_path)
 
+    def test_update_queue(self):
+        queue1_path = "%s/queues/test_queue1" % self._parent
+
+        ret = self._client.update_queue(
+            {"name": queue1_path}
+        )
+        self.assertEqual(ret.name, queue1_path)
+
+        queue1_path = "%s/queues/test_queue1" % self._parent
+
+        ret = self._client.update_queue(
+            {"name": queue1_path}, update_mask={}
+        )
+        self.assertEqual(ret.name, queue1_path)
+
     def test_list_queues(self):
         path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
         self._client.delete_queue(path)
@@ -135,6 +150,20 @@ class TestCase(BaseTestCase):
         response = self._client.pause_queue(path)
         self.assertEqual(response.state, 2)
 
+    def test_resume_queue(self):
+        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "default")
+        self._client.delete_queue(path)
+
+        self.test_create_queue()  # Create a couple of queues
+
+        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
+        response = self._client.pause_queue(path)
+        self.assertEqual(response.state, 2)
+
+        path = self._client.queue_path('[PROJECT]', '[LOCATION]', "test_queue2")
+        response = self._client.resume_queue(path)
+        self.assertEqual(response.state, 1)
+
     def test_purge_queue(self):
         self.test_create_queue()  # Create a couple of queues
 
